#!/usr/bin/env python
import datetime
import os.path

import dcdr.data as dt
import dcdr.load.xmlspec as xmlspec
import dcdr.output.instance as inst
import dcdr.sequence as seq

class FatError(Exception):
    pass

class _DirectoryEntry:
    """
    A class holding information on an entry in a directory.
    """
    def __init__(self, filename, attributes, create_timestamp, access_timestamp, modified_timestamp, cluster, size, file_allocation_table, is_fat_32):
        self.filename = filename
        self.attributes = attributes
        self.create_timestamp = create_timestamp
        self.access_timestamp = access_timestamp
        self.modified_timestamp = modified_timestamp
        self.cluster = cluster
        self.size = size

        self._file_allocation_table = file_allocation_table
        self._cluster_set = set()
        self._cluster_chain = []
        self._is_fat_32 = is_fat_32

    def __str__(self):
        if self.attributes.subdirectory:
            result = "[d] %s" % self.filename
        else:
            result = "%s (%i bytes)" % (self.filename, self.size)
        return result

    def _get_fat_info(self, cluster):
        """
        Get the cluster information from the fat (eg: next cluster, unused, ...)
        """
        # Note that the first two clusters are reserved, and cannot be
        # referenced in a chain.
        if cluster >= len(self._file_allocation_table) or cluster < 2:
            raise FatError('Asked file allocation table for status of out of bounds cluster %i (max %i)' % (cluster, len(self._file_allocation_table)))

        return self._file_allocation_table[cluster]

    def _walk_cluster_chain(self, cluster):
        """
        Iterate through a cluster chain given a starting cluster.
        """
        if self._is_fat_32:
            BAD_CLUSTER = 0x0FFFFFF7
            FINAL_CLUSTER = 0x0FFFFFFF
            MASK = 0x0FFFFFFF
        else:
            BAD_CLUSTER = 0x0FF7
            FINAL_CLUSTER = 0x0FFF
            MASK = 0x0FFF

        while 1:
            # Lets have a look at the FAT and make sure this cluster is actually allocated...
            next_cluster = self._get_fat_info(cluster)
            if next_cluster & MASK == 0x0:
                raise FatError("Freed cluster in chain for entry '%s' (starting at %i)!" % (self.filename, self.cluster))

            if next_cluster & MASK == BAD_CLUSTER:
                raise FatError("Bad cluster in chain for entry '%s' (starting at %i)!" % (self.filename, self.cluster))

            yield cluster

            if next_cluster & MASK == FINAL_CLUSTER:
                # No more clusters in the chain
                break

            cluster = next_cluster

    def get_cluster_chain(self, index=0):
        """
        Query the cluster chain for the given directory entry.

        Takes an index to the cluster in the chain to start at.
        """
        # Clusters are a 'linked list' in the file allocation table. As typical
        # access is to randomly access a specific cluster, we keep a cache of
        # the clusters in a list for fast access.
        if len(self._cluster_chain) > 0:
            i = min(index, len(self._cluster_chain) - 1)
            cluster = self._cluster_chain[i]
        else:
            i = 0
            cluster = self.cluster

        for offset, cluster in enumerate(self._walk_cluster_chain(cluster)):
            if i + offset == len(self._cluster_chain):
                # Add this cluster to the cache so we can access it quickly
                # next time.
                if cluster in self._cluster_set:
                    raise FatError("Recursive cluster %i in chain for entry '%s' (starting at %i)!" % (cluster, self.filename, self.cluster))
                self._cluster_set.add(cluster)
                self._cluster_chain.append(cluster)

            if i + offset >= index:
                yield cluster

class Fat:
    """
    Class providing high level file system routines from a FAT16/FAT32 filesystem image.
    """
    BYTES_PER_DIRECTORY_ENTRY = 32

    def __init__(self, image):
        if os.path.splitext(image)[1] == ".gz":
            import gzip
            self._file = gzip.open(image, 'rb')
        else:
            self._file = file(image, 'rb')
        
        # Load the boot sector
        entry = xmlspec.Importer().load(file('bootsector.xml', 'r'))
        result = inst.decode(entry, dt.Data(self._get_buffer(0, 512)))
        self._bootsector = result.fat_bootsector
        self._dir_cache = {}
        self._dir_entry_decoder = xmlspec.Importer().load(file('filename.xml', 'r'))
        self._file_allocation_table = self._load_fat()

        # Setup the root directory entry
        class _RootAttributes:
            def __init__(self):
                self.subdirectory = 1
        timestamp = datetime.datetime.fromtimestamp(0)

        if self._is_fat_32():
            root_cluster = self._get_boot_sector().cluster_number_of_root_directory_start
        else:
            root_cluster = 0
        self._root_directory = _DirectoryEntry('', _RootAttributes(), timestamp, timestamp, timestamp, root_cluster, 0, self._file_allocation_table, self._is_fat_32())

    def _load_fat(self):
        """
        Load a decoded copy of the file allocation table.
        
        Returns a list of integers, where result[cluster] is the FAT status
        for the given cluster.
        """
        if self._is_fat_32():
            bytes_per_entry = 4
        else:
            bytes_per_entry = 2

        boot = self._get_boot_sector()
        fat_offset = boot.reserved_sector_count * boot.bytes_per_sector
        fat_size = boot.sectors_per_fat * boot.bytes_per_sector
        buffer = self._get_buffer(fat_offset, fat_offset + fat_size)

        num_clusters = fat_size / bytes_per_entry
        data = dt.Data(buffer)
        return list(data.pop(bytes_per_entry * 8).get_little_endian_integer() for i in range(num_clusters))

    def _get_buffer(self, start, end):
        """ Get a buffer for sequence of bytes in range [start, end) """
        self._file.seek(start)
        buffer = ""
        size = end - start
        while len(buffer) != size:
            data = self._file.read(size - len(buffer))
            if len(data) == 0:
                raise Exception("Not enough data (needs %i bytes, got %i bytes)" % (end, len(buffer)))
            buffer += data
        return buffer

    def _is_fat_32(self):
        return self._get_boot_sector().fat_file_system_type == "FAT32   "

    def _get_boot_sector(self):
        return self._bootsector

    def _get_checksum(self, filename):
        """
        Get the FAT checksum of a given filename.

        See 'http://en.wikipedia.org/wiki/File_Allocation_Table'.
        """
        sum = 0
        for char in filename:
            if sum & 1:
                sum = 0x80 + (sum >> 1)
            else:
                sum = (sum >> 1)
            sum += ord(char)
            sum &= 0xff
        return sum
    
    def _list_directory(self, binary):
        """ Get an iterator to the contents of a given directory """
        assert len(binary) % self.BYTES_PER_DIRECTORY_ENTRY == 0
        num_entries = len(binary) / self.BYTES_PER_DIRECTORY_ENTRY
        long_filename = None
        checksum = None
        data = dt.Data(binary)
        for i in range(num_entries):
            entry = inst.decode(self._dir_entry_decoder, data)
            if hasattr(entry.file_entry, "no_more_entries"):
                # There are no more items in this directory.
                break

            if hasattr(entry.file_entry, "dos_filename"):
                df = entry.file_entry.dos_filename

                filename = df.filename.strip()
                if df.lowercase_basename:
                    filename = filename.lower()

                ext = df.extension.strip()
                if df.lowercase_extension:
                    ext = ext.lower()
                if ext:
                    filename = "%s.%s" % (filename, ext)

                if long_filename is not None:
                    # We have a long filename available; use it if
                    # the checksum matches
                    if self._get_checksum(df.filename + df.extension) == checksum:
                        filename = long_filename
                long_filename = None 
                cluster = df.first_cluster
                if self._is_fat_32():
                    cluster = (df.ea_index << 16) + df.first_cluster
                
                yield _DirectoryEntry(
                        filename,
                        df.attributes,
                        self._get_timestamp(df.create_timestamp),
                        self._get_timestamp(df.access_timestamp),
                        self._get_timestamp(df.modified_timestamp),
                        cluster,
                        df.file_size,
                        self._file_allocation_table,
                        self._is_fat_32())
            elif hasattr(entry.file_entry, "long_filename"):
                lf = entry.file_entry.long_filename
                filename = lf.chars_1 + lf.chars_2 + lf.chars_3
                if '\x00' in filename:
                    filename = filename[:filename.index('\x00')]

                if long_filename is None or lf.checksum != checksum:
                    long_filename = filename
                    checksum = lf.checksum
                else:
                    long_filename = filename + long_filename

    def _get_timestamp(self, ts):
        microsecond = 0
        second = 0
        if hasattr(ts, 'centiseconds'):
            microsecond = ts.centiseconds * 10000

            # The 'centiseconds' stores two seconds...
            if microsecond >= 1000000:
                second = 1
                microsecond -= 1000000

        hour = minute = 0
        if hasattr(ts, 'hour'):
            hour = ts.hour
            minute = ts.minute_upper * 8 + ts.minute_lower
            second += ts.two_seconds * 2
            
        return datetime.datetime(ts.year + 1980, ts.month_upper * 8 + ts.month_lower, ts.day, hour, minute, second, microsecond)

    def listdir(self, entry):
        """
        Return the contents of a directory
        """
        assert entry.attributes.subdirectory

        entries = self._dir_cache.get(entry.cluster)
        if entries is not None:
            return entries

        if entry.cluster == 0:
            # This is the fat16 root directory
            boot = self._get_boot_sector()
            root_dir_sector = boot.reserved_sector_count + boot.sectors_per_fat * boot.number_of_file_allocation_tables
 
            offset = root_dir_sector * boot.bytes_per_sector
            binary = self._get_buffer(offset, offset + boot.max_root_directory_entries * self.BYTES_PER_DIRECTORY_ENTRY)
        else:
            # Request the complete cluster chain
            binary = "".join(self._get_cluster_contents(cluster) for cluster in entry.get_cluster_chain())

        entries = list(self._list_directory(binary))
        self._dir_cache[entry.cluster] = entries
        return entries

    def read(self, entry, offset, length):
        """
        Read the contents of a file
        """
        assert not entry.attributes.subdirectory

        boot = self._get_boot_sector()
        bytes_per_cluster = boot.sectors_per_cluster * boot.bytes_per_sector

        data = ""
        length_remaining = length

        start_cluster_index = offset / bytes_per_cluster
        offset -= start_cluster_index * bytes_per_cluster

        for cluster in entry.get_cluster_chain(start_cluster_index):
            cluster_length = min(length_remaining, bytes_per_cluster - offset)
            length_remaining -= cluster_length
            assert length_remaining >= 0

            data = data + self._get_cluster_contents(cluster)[offset:offset + cluster_length]
            offset = 0

            if length_remaining == 0:
                break

        assert len(data) == length, "%i == %i" % (len(data), length)
        return data

    def get_root_dir(self):
        """
        Get the root directory entry object
        """
        return self._root_directory

    def _get_cluster_contents(self, cluster):
        """ Get the contents of a given cluster """
        boot = self._get_boot_sector()
        num_clusters = boot.total_sectors / boot.sectors_per_cluster
        if cluster < 2 or cluster >= num_clusters:
            raise FatError("Requested contents of out of bounds cluster %i [2, %i)!" % (cluster, num_clusters))

        root_dir_sector = boot.reserved_sector_count + boot.sectors_per_fat * boot.number_of_file_allocation_tables
        cluster_origin = root_dir_sector * boot.bytes_per_sector + boot.max_root_directory_entries * self.BYTES_PER_DIRECTORY_ENTRY
        offset = (cluster - 2) * boot.sectors_per_cluster * boot.bytes_per_sector + cluster_origin
        return self._get_buffer(offset, offset + boot.sectors_per_cluster * boot.bytes_per_sector)


if __name__ == "__main__":
    import sys
    fat = Fat(sys.argv[1])
    for entry in fat.listdir(fat.get_root_dir()):
        print entry
